/* *********************************************************

class TUtil : Utility class for ROOT
  author    : Eito IWAI ( iwai # champ.hep.sci.osaka-u.ac.jp )

history :
   v0.4.2   enable to deal with the recent g++ compiler
   v0.4.1   add TUtil::DrawNormalized method
   v0.4.0   enable to choose whether making output files in default constructor
   v0.3.3   add TUtil::SetStyleTable method
            replace some char* to const char*
   v0.3.2   add TUtil::ReOpen() method to activate the output ROOT file
            add gStyle->SetFrameFillColor(0) to init_style().
   v0.3.1   add some handlings in the constructor for the odd behavior with CINT
   v0.3.0   enable to treat TProfile with TUtil::Draw
            remove m_drawObj.clear() in destructor
            change TUtil::(Get)Version() method to a static function
	    add TMath.h to include header file
   v0.2.7   check the validity of the superimpose option
            a overlaid color processing is modified
   v0.2.6   add TUtil::NewTGraphAsymmErrors method
            change the timing of deleting clone objects in destructor
   v0.2.5   rename the output file ( *.ps, *.root ), if already exists
   v0.2.4   delete clone objects in destructor
   v0.2.3   add a flag m_applyColorToFirstDraw and also its setter method
            update the color processes on overlaid drawing
            check the pointer to an drawed object in skeletonDraw
   v0.2.2   fix a bug with TUtil::Fit( ..., TF1* ...)
   v0.2.1   futher more update draw methods, especially TUtil::Fit
            update SetPadNumXY
   v0.2.0   change the treatment of draw objects ( make clone )
            change the specification & usage for TUtil::Draw(TTree... )
            fix a bug with m_colorTable
   v0.1.7   add "#define TUtil_h" to avoid redefinitions
            change private to protected
            add const statement with extractor
   v0.1.6   add TFile::Clear() after Write()
   v0.1.5b  change the treatment of TGraph/TGraphErrors
   v0.1.5a  NewTGraph/NewTGraphErrors are modified ( add skeletonNewTGraph )
            add the certain process to store TGraphs to ROOT file
   v0.1.4   Fix red-line shade around draw area ( Thanks to Nakaya )
            NewTGraph and NewTGraphErrors are modified
   v0.1.3   Draw and Fit are modified ( update skeletonDraw )
            Draw(TTree*) is added
            ViewHists is added
   v0.1.2   Update and CdPad are modified
            SetPadNum,SetPadNumX,SetPadNumY and SetPadNumXY are added
            Draw is modified ( add skeletonDraw )
            remove ClassDef/ClassImp temporary?
   v0.1.1   NewTGraphErrors and Draw(TGraph*) are added
            ClassDef/ClassImp, virtualize destructor are added
   v0.1.0   first version
   
********************************************************** */

#ifndef TUtil_h
#define TUtil_h

#define TUTIL_PSFILE "t.ps"
#define TUTIL_ROOTFILE "t.root"
#define TUTIL_VERSION "TUtil v0.4.2_20090425"

#include <iostream>
#include <vector>
#include <string>
#include <sstream>

#include <TH1.h>
#include <TCanvas.h>
#include <TPostScript.h>
#include <TSystem.h>
#include <TFile.h>
#include <TStyle.h>
#include <TROOT.h>
#include <TGraph.h>
#include <TGraphErrors.h>
#include <TGraphAsymmErrors.h>
#include <TTree.h>
#include <TKey.h>
#include <TCollection.h>
#include <TF1.h>
#include <TMath.h>

class TUtil{
public:
  TUtil(bool psFile=kTRUE, bool rootFile=kTRUE);
  virtual ~TUtil(void);
  
  // Pad Util
  void CdPad(void);
  void cdPad(void){ CdPad(); }
  void Update(int padNum=1);
  
  // Draw
  void Draw(TH1* hist, const char* option="", int logScale=000){ skeletonDraw(hist,option,logScale); }
  void Draw(TH1& hist, const char* option="", int logScale=000){ Draw(&hist,option,logScale); }
  void Draw(const char* histName, const char* option="", int logScale=000){ Draw((TH1*)gROOT->FindObject(histName),option,logScale); }
  void Draw(TH1* hist, int logScale, const char* option=""){ Draw(hist,option,logScale); }
  void Draw(TH1& hist, int logScale, const char* option=""){ Draw(&hist,option,logScale); }
  void Draw(const char* histName, int logScale, const char* option=""){ Draw((TH1*)gROOT->FindObject(histName),option,logScale); }
  
  void Draw(TGraph* graph, const char* option="", int logScale=000){ skeletonDraw(graph,option,logScale); }
  void Draw(TGraph& graph, const char* option="", int logScale=000){ Draw(&graph,option,logScale); }
  
  TH1* Draw(TTree* tree, const char* varexp, const char* selection="", const char* option="", int logScale=000);
  TH1* Draw(TTree& tree, const char* varexp, const char* selection="", const char* option="", int logScale=000){ return Draw(&tree,varexp,selection,option,logScale); }
  
  void DrawNormalized(TH1* hist, const char* option="", int logScale=000){ skeletonDraw(hist,option,logScale,3); }
  void DrawNormalized(TH1& hist, const char* option="", int logScale=000){ DrawNormalized(&hist,option,logScale); }
  void DrawNormalized(const char* histName, const char* option="", int logScale=000){ DrawNormalized((TH1*)gROOT->FindObject(histName),option,logScale); }
  void DrawNormalized(TH1* hist, int logScale, const char* option=""){ DrawNormalized(hist,option,logScale); }
  void DrawNormalized(TH1& hist, int logScale, const char* option=""){ DrawNormalized(&hist,option,logScale); }
  void DrawNormalized(const char* histName, int logScale, const char* option=""){ DrawNormalized((TH1*)gROOT->FindObject(histName),option,logScale); }
  
  // Fit
  void Fit(TH1* hist, char* formula, const char* option="", int logScale=000, const char* goption="", double xmin=0, double xmax=0){ skeletonDraw(hist,goption,logScale,1,formula,option,xmin,xmax); }
  void Fit(TH1& hist, char* formula, const char* option="", int logScale=000, const char* goption="", double xmin=0, double xmax=0){ Fit(&hist,formula,option,logScale,goption,xmin,xmax); }
  void Fit(const char* histName, char* formula, const char* option="", int logScale=000, const char* goption="", double xmin=0, double xmax=0){ Fit((TH1*)gROOT->FindObject(histName),formula,option,logScale,goption,xmin,xmax); }
  void Fit(TH1* hist, TF1* f1, const char* option="", int logScale=000, const char* goption="", double xmin=0, double xmax=0){ skeletonDraw(hist,goption,logScale,2,f1,option,xmin,xmax); }
  void Fit(TH1& hist, TF1* f1, const char* option="", int logScale=000, const char* goption="", double xmin=0, double xmax=0){ Fit(&hist,f1,option,logScale,goption,xmin,xmax); }
  void Fit(const char* histName, TF1* f1, const char* option="", int logScale=000, const char* goption="", double xmin=0, double xmax=0){ Fit((TH1*)gROOT->FindObject(histName),f1,option,logScale,goption,xmin,xmax); }
  
  // new blank TGraph ( to be stored to ROOT file )
  TGraph* NewTGraph(char* name="defaultName"){ return skeletonNewTGraph(0,name); }
  TGraphErrors* NewTGraphErrors(char* name="defaultName"){ return (TGraphErrors*)skeletonNewTGraph(1,name); }
  TGraphAsymmErrors* NewTGraphAsymmErrors(char* name="defaultName"){ return (TGraphAsymmErrors*)skeletonNewTGraph(2,name); }
  
  void ViewHists(TFile* f, const char* searchKey="", TString::ECaseCompare cmp=TString::kIgnoreCase);
  void ReOpen(char* mode="UPDATE");
  
  // setter
  void SetPadNum(int padNum){ m_padNum=padNum; }
  void SetPadNumX(int padNumX){ SetPadNumXY(padNumX,m_padNumY); }
  void SetPadNumY(int padNumY){ SetPadNumXY(m_padNumX,padNumY); }
  void SetPadNumXY(int padNumX, int padNumY);
  //void SetCanvas(TCanvas& c);
  void SetColorTable(std::vector<int> colorTable){ m_colorTable=colorTable; }
  void SetStyleTable(std::vector<int> styleTable){ m_styleTable=styleTable; }
  void SetApplyColorToFirstDraw(bool applyColor=true){ m_applyColorToFirstDraw=applyColor; }
  
  // extractor, getter
  int GetPadNum(void) const { return m_padNum; }
  int GetPadNumX(void) const { return m_padNumX; }
  int GetPadNumY(void) const { return m_padNumY; }
  TCanvas* GetCanvas(void) const { return m_c; }
  
  // static function
  static const std::string GetVersion(void){ return TUTIL_VERSION; }
  static const std::string Version(void){ return GetVersion(); }
  
protected:
  
  // init
  void init_style(void);
  void init_canvas(void);
  void init_ps(void);
  void init_ofile(void);
  void init_colortable(void);
  
  // skeleton
  void skeletonDraw(void* obj, const char* option, int logScale, int drawType=0, void* fitFunc=0, const char* fitOption="", double xmin=0., double xmax=0.);
  TGraph* skeletonNewTGraph(int graphType, const char* name);
  
  bool m_bps;
  bool m_broot;
  int m_padNum;
  int m_padNumX;
  int m_padNumY;
  TCanvas* m_c;
  TPostScript* m_ps;
  TFile* m_f;
  int m_colorTableNum;
  std::vector<int> m_colorTable;
  std::vector<int> m_styleTable;
  std::vector<TObject*> m_drawObj;
  bool m_applyColorToFirstDraw;
  
  //ClassDef(TUtil,1)
};

//#if !defined(__CINT__)
//ClassImp(TUtil)
//#endif

// constructor
TUtil::TUtil(bool psFile, bool rootFile):m_bps(psFile),m_broot(rootFile),m_padNum(0),m_padNumX(2),m_padNumY(3),m_colorTableNum(0),m_applyColorToFirstDraw(1){
  init_style();
  init_canvas();
  if ( m_bps ){ init_ps(); }
  if ( m_broot ){ init_ofile(); }
  init_colortable();
  
  // for the odd behavior in interactive/macro mode (CINT bug??)
#ifdef __CINT__
  m_drawObj.push_back(new TObject);
  delete m_drawObj[0];
  m_drawObj.clear();
#endif // __CINT__
  
}

// destructor
TUtil::~TUtil(void){
  if ( m_bps ){ m_ps->Close(); }
  for ( int i=0;i<(int)m_drawObj.size();i++ ){ delete m_drawObj[i]; }
  if ( m_broot ){ m_f->Write(); m_f->Clear(); }
  if ( m_bps ){ delete m_ps; }
  if ( m_broot ){ delete m_f; }
  delete m_c;
}

void TUtil::CdPad(void){
  m_c->Update();
  m_padNum++;
  if ( m_padNum==m_padNumX*m_padNumY+1 ){
    Update();
    m_padNum=1;
  }
  m_c->cd(m_padNum);
  m_colorTableNum=0;
}

void TUtil::Update(int padNum){
  if ( !(1<=padNum && padNum<=m_padNumX*m_padNumY) ){
    std::cerr << "[Error] TUtil::Update >> padNum is out of range, ignored." << std::endl;
    return;
  }
  m_c->Update();
  
  for ( int i=0;i<(int)m_drawObj.size();i++ ){ delete m_drawObj[i]; }
  m_drawObj.clear();
  
  if ( m_padNum!=0 && m_bps ){ m_ps->NewPage(); }
  for ( int i=0;i<m_padNumX*m_padNumY;i++ ){
    m_c->cd(i+1);
    gPad->SetLogx(0);
    gPad->SetLogy(0);
    gPad->SetLogz(0);
    gPad->Clear();
  }
  m_c->cd(padNum);
  m_padNum=padNum-1;
}

// init
void TUtil::init_canvas(void){
  m_c=new TCanvas("c","ROOT canvas",0,0,640,856);
  m_c->Divide(m_padNumX, m_padNumY, 0.000001, 0.000001);
  
  int i=1;
  while( i<=m_padNumX*m_padNumY ){
    m_c->cd(i);
    gPad->SetGridx();
    gPad->SetGridy();
    i++;
  }
  m_c->cd(1);
}

void TUtil::init_ps(void){
  if ( !gSystem->AccessPathName(TUTIL_PSFILE) ){
    std::string t_bkpName=TUTIL_PSFILE;
    gSystem->Rename(TUTIL_PSFILE,(t_bkpName+"~").c_str());
  }
  m_ps=new TPostScript(TUTIL_PSFILE,111);
  m_ps->Range(20,40);
  m_ps->NewPage();
}

void TUtil::init_ofile(void){
  if ( !gSystem->AccessPathName(TUTIL_ROOTFILE) ){
    std::string t_bkpName=TUTIL_ROOTFILE;
    gSystem->Rename(TUTIL_ROOTFILE,(t_bkpName+"~").c_str());
  }
  m_f=new TFile(TUTIL_ROOTFILE,"RECREATE");
}

void TUtil::init_colortable(void){
  m_colorTable.push_back(kBlack);
  m_colorTable.push_back(kRed);
  m_colorTable.push_back(kBlue);
  m_colorTable.push_back(kMagenta);
  
  m_colorTable.push_back(kSpring-1);
  m_colorTable.push_back(kAzure-3);
  m_colorTable.push_back(kOrange+1);
  
  m_styleTable.push_back(1);
  m_styleTable.push_back(2);
}

void TUtil::init_style(void){
  gStyle->SetStatColor(0);
  gStyle->SetTitleFillColor(0);
  gStyle->SetTitleH(0.05);
  gStyle->SetTitleBorderSize(1);
  gStyle->SetStatX(0.84);
  gStyle->SetStatY(0.88);
  gStyle->SetStatFontSize(0.023);
  gStyle->SetStatBorderSize(1);
  gStyle->SetStatStyle(0);
  gStyle->SetFrameFillColor(0);
  gStyle->SetFrameBorderSize(0);
  gStyle->SetFrameBorderMode(0);
  gStyle->SetLegendBorderSize(1);
  gStyle->SetCanvasColor(0);
  gStyle->SetPadTopMargin(0.12);
  gStyle->SetPadBottomMargin(0.12);
  gStyle->SetPadRightMargin(0.14);
  gStyle->SetPadLeftMargin(0.12);
  gStyle->SetLabelSize(0.05,"X");
  gStyle->SetLabelSize(0.05,"Y");
  gStyle->SetMarkerStyle(20);
  gStyle->SetMarkerSize(0.6);
  gStyle->SetFuncWidth(1);
  gStyle->SetFuncColor(2);
  gStyle->SetEndErrorSize(0);
  gStyle->SetNdivisions(505);
  gStyle->SetPalette(1);
  
  gStyle->SetOptStat("neMri");
  gStyle->SetOptFit(0001);
  gStyle->SetStatW(0.3);
  gStyle->SetStatFontSize(0.03);
}

// Draw
void TUtil::skeletonDraw(void* obj, const char* option, int logScale, int drawType, void* fitFunc, const char* fitOption, double xmin, double xmax){
  
  if ( obj==NULL ){
    std::cerr << "[Error] TUtil::Draw(...) >> Invalid argument. The pointer to a draw object is NULL, ignored." << std::endl;
    return;
  }
  
  TString t_opt=option;
  bool b_same=kFALSE;
  TString t_classname=((TObject*)obj)->ClassName();
  if ( t_classname.Contains("TGraph") ){
    if ( !t_opt.Contains("A",TString::kIgnoreCase) ){ b_same=kTRUE; }
    if ( b_same && gPad->GetListOfPrimitives()->GetEntries()==0 ){
      t_opt.Append("A");
      b_same=kFALSE;
      std::cerr << "[Error] TUtil::Draw(...) >> Missing option 'A', automatically added." << std::endl;
    }
  }else if ( t_classname.Contains("TH") || t_classname.Contains("TProfile") ){
    if ( t_opt.Contains("same",TString::kIgnoreCase) ){ b_same=kTRUE; }
    if ( b_same && gPad->GetListOfPrimitives()->GetEntries()==0 ){
      if ( t_opt.Contains("sames",TString::kIgnoreCase) ){
	t_opt.Remove(t_opt.Index("sames",0,TString::kIgnoreCase),5);
      }else{
	t_opt.Remove(t_opt.Index("same",0,TString::kIgnoreCase),4);
      }
      b_same=kFALSE;
      std::cerr << "[Error] TUtil::Draw(...) >> Invalid option 'SAME(S)', ignored." << std::endl;
    }
  }
  
  if ( !b_same ){
    CdPad();
    bool b_log[3];
    b_log[2]=logScale%2;
    b_log[1]=(logScale%100>2) ? 1 : 0;
    b_log[0]=(logScale>20) ? 1 : 0;
    gPad->SetLogx(b_log[0]);
    gPad->SetLogy(b_log[1]);
    gPad->SetLogz(b_log[2]);
  }
  
  std::string t_fitOption=fitOption;
  switch(drawType){
  case 0: // Draw
    break;
  case 1: // Fit(char*, ...)
    ((TH1*)obj)->Fit((char*)fitFunc,(t_fitOption+'0').c_str(),"",xmin,xmax);
    ((TH1*)obj)->GetFunction((const char*)fitFunc)->ResetBit(1<<9);
    break;
  case 2: // Fit(TF1*, ...)
    ((TH1*)obj)->Fit((TF1*)fitFunc,(t_fitOption+'0').c_str(),"",xmin,xmax);
    ((TH1*)obj)->GetFunction(((TF1*)fitFunc)->GetName())->ResetBit(1<<9);
    break;
  case 3: // DrawNormalized
    if ( ((TH1*)obj)->GetSumOfWeights()==0 ){
      std::cerr << "[Error] TUtil::skeletonDraw >> Sum of weights is null, ignored." << std::endl;
      return;
    }
    //((TH1*)obj)->Sumw2();
    ((TH1*)obj)->Scale(1./((TH1*)obj)->GetSumOfWeights());
    break;
  default:
    std::cerr << "[Error] TUtil::skeletonDraw >> Invalid drawType, ignored." << std::endl;
    return;
  }
  
  TAttLine* t_line=0;
  TAttMarker* t_marker=0;
  void* drawObj=0;
  if ( t_classname.Contains("TGraph") ){
    drawObj=((TGraph*)obj)->Clone();
    m_drawObj.push_back((TGraph*)drawObj);
    t_line=(TGraph*)drawObj;
    t_marker=(TGraph*)drawObj;
  }else if ( t_classname.Contains("TH") || t_classname.Contains("TProfile") ){
    drawObj=((TH1*)obj)->Clone();
    m_drawObj.push_back((TH1*)drawObj);
    t_line=(TH1*)drawObj;
    t_marker=(TH1*)drawObj;
  }else{
    std::cerr << "[Error] TUtil::skeletonDraw >> Unknown object type, ignored." << std::endl;
    return;
  }
  ((TNamed*)drawObj)->Draw(t_opt);
  
  if ( b_same ){
    if ( m_colorTableNum==0 && m_applyColorToFirstDraw && m_drawObj.size()>=2 ){
      TAttLine* t_firstDrawLine=0;
      TAttMarker* t_firstDrawMarker=0;
      TString t_firstDrawClassName=m_drawObj[m_drawObj.size()-2]->ClassName();
      if ( t_firstDrawClassName.Contains("TGraph") ){
	t_firstDrawLine=(TGraph*)m_drawObj[m_drawObj.size()-2];
	t_firstDrawMarker=(TGraph*)m_drawObj[m_drawObj.size()-2];
      }else if ( t_firstDrawClassName.Contains("TH") || t_firstDrawClassName.Contains("TProfile") ){
	t_firstDrawLine=(TH1*)m_drawObj[m_drawObj.size()-2];
	t_firstDrawMarker=(TH1*)m_drawObj[m_drawObj.size()-2];
      }
      t_firstDrawLine->SetLineColor(m_colorTable[0]);
      t_firstDrawMarker->SetMarkerColor(m_colorTable[0]);
    }
    m_colorTableNum++;
    if ( m_colorTableNum>=(int)m_colorTable.size()*(int)m_styleTable.size() ){ m_colorTableNum=0; }
    if ( m_colorTableNum<(int)m_colorTable.size() ){
      t_line->SetLineColor(m_colorTable[m_colorTableNum]);
      t_marker->SetMarkerColor(m_colorTable[m_colorTableNum]);
    }else{
      t_line->SetLineStyle(m_styleTable[m_colorTableNum/m_colorTable.size()]);
      t_line->SetLineColor(m_colorTable[m_colorTableNum%m_colorTable.size()]);
      t_marker->SetMarkerColor(m_colorTable[m_colorTableNum%m_colorTable.size()]);
    }
  }
  
  //if ( !b_same ){ m_c->Update(); }
}

TH1* TUtil::Draw(TTree* tree, const char* varexp, const char* selection, const char* option, int logScale){
  
  TString t_varexp=varexp;
  std::string t_histname="htemp";
  if ( t_varexp.Contains(">>") ){
    t_varexp.Remove(0,t_varexp.Index(">>")+2);
    if ( t_varexp.Index("(")>=0 ){ t_varexp.Remove(t_varexp.Index("(")); }
    t_varexp=t_varexp.Strip(TString::kBoth);
    t_histname=t_varexp;
  }
  
  TString t_option=option;
  t_option.Append("goff");
  tree->Draw(varexp,selection,t_option);
  TH1* t_hist=(TH1*)(gROOT->FindObject(t_histname.c_str()));
  t_hist->SetName((" "+t_histname+" ").c_str());
  
  skeletonDraw(t_hist,option,logScale,0);
  
  return t_hist;
}

TGraph* TUtil::skeletonNewTGraph(int graphType, const char* name){
  TGraph* t_gr=NULL;
  switch(graphType){
  case 0: // TGraph
    t_gr=new TGraph;
    break;
  case 1: // TGraphErrors
    t_gr=new TGraphErrors;
    break;
  case 2: // TGraphAsymmErrors
    t_gr=new TGraphAsymmErrors;
    break;
  default:
    std::cerr << "[Error] TUtil::skeletonNewTGraph >> Invalid graphType, ignored." << std::endl;
  }
  
  gDirectory->Append(t_gr);
  if ( TString(name)=="defaultName" ){ t_gr->SetName("Graph"); }else{ t_gr->SetName(name); }
  
  return t_gr;
}

void TUtil::SetPadNumXY(int padNumX, int padNumY){
  Update();
  if ( m_bps ){ m_ps->Off(); }
  delete m_c;
  m_padNumX=padNumX;
  m_padNumY=padNumY;
  init_canvas();
  if ( m_bps ){ m_ps->On(); }
}

void TUtil::ViewHists(TFile* f, const char* searchKey, TString::ECaseCompare cmp){
  TIter t_iter(f->GetListOfKeys());
  TKey* t_key;
  while ( (t_key=(TKey*)t_iter.Next()) ){
    TString t_classname=t_key->GetClassName();
    if ( !t_classname.Contains("TH") ){ continue; }
    TString t_name=t_key->GetName();
    if ( !t_name.Contains(searchKey,cmp) ){ continue; }
    TH1* t_hist=(TH1*)(gROOT->FindObject(t_name));
    Draw(t_hist);
  }
}

void TUtil::ReOpen(char* mode){
  if ( !m_broot ){ std::cerr << "[Error] TUtil::ReOpen >> ROOT-file was not created, ignored." << std::endl; return; }
  m_f->ReOpen(mode);
}

#endif // TUtil_h
